class AppNoRowsFound(Exception):
    code = 404


class AppException(Exception):
    message = "Error"
    code = 500
    custom_exception = True
    silent = False
    use_default_response_message_key = True


class AppSilentException(AppException):
    silent = True


class AppNotSilentException(AppException):
    silent = False


class AppValidationError(AppSilentException):
    code = 400
    message = "Validation Error"


class AppNotSilentValidationError(AppNotSilentException):
    code = 400
    message = "Not Silent Validation Error"


class AppExpiredTokenError(AppSilentException):
    code = 401
    message = "Token Error"


class AppAccessTokenExpiredError(AppSilentException):
    code = 419
    message = "Access token expired"


class AppRefreshTokenInvalidError(AppSilentException):
    code = 401
    message = "Refresh token is invalid"


class AppMicroFetcherError(AppSilentException):
    code = 400
    message = "Fetcher Service Error"


class AppThirdPartyError(AppNotSilentException):
    code = 400
    message = "Third Party Call Error"


class AppFetchServiceDataError(AppNotSilentException):
    code = 400
    message = "Error while fetch relational data from service"

    def __init__(self, message=None, code=None):
        if message:
            self.message = message
        if code:
            self.code = code


class AppNotFoundError(AppSilentException):
    code = 404
    message = "Not found"


class AppMissingAuthError(AppSilentException):
    code = 401
    message = "You are unauthenticated"


class AppDeprecatedApiError(AppSilentException):
    code = 410
    message = "This API is deprecated and no longer available"


class AppForbiddenError(AppSilentException):
    code = 403
    message = "You are not able to use this API"


class AppPreconditionNotMetError(AppSilentException):
    code = 409
    message = "Action not allowed. Please meet all preconditions."


class CustomNotFoundError(AppSilentException):
    def __init__(self, message):
        self.message = message

    code = 404


class AppTokenRefreshRequired(Exception):
    """Exception raised when a token refresh is required and a new access token is available."""
    def __init__(self, message, new_access_token=None):
        self.message = message
        self.new_access_token = new_access_token
        super().__init__(self.message)



from wedeliver_core_plus import WedeliverCorePlus
from wedeliver_core_plus.helpers.auth import Auth
from sqlalchemy import event


def init_base_model():
    app = WedeliverCorePlus.get_app()
    db = app.extensions['sqlalchemy'].db

    class BaseModel(db.Model):
        __abstract__ = True
        id = db.Column(db.Integer, primary_key=True)
        creation = db.Column(db.DateTime, default=db.func.now())
        created_by = db.Column(db.String(64), default=Auth.get_user_str)

        modification = db.Column(db.DateTime, default=db.func.now(), onupdate=db.func.now())

        def __init__(self, *args, **kwargs):
            super(BaseModel, self).__init__(*args, **kwargs)
            self.register_hooks()

        # def __init__(self, created_by=None):
        #     self.created_by = created_by or Auth.get_user().get("email", "Guest")

        @classmethod
        def register_hooks(cls):
            @event.listens_for(cls, 'before_insert')
            def before_insert(mapper, connection, target):
                if hasattr(target, 'before_insert'):
                    target.before_insert(mapper, connection)

            @event.listens_for(cls, 'after_insert')
            def after_insert(mapper, connection, target):
                if hasattr(target, 'after_insert'):
                    target.after_insert(mapper, connection)

            @event.listens_for(cls, 'before_update')
            def before_update(mapper, connection, target):
                if hasattr(target, 'before_update'):
                    target.before_update(mapper, connection)

            @event.listens_for(cls, 'after_update')
            def after_update(mapper, connection, target):
                if hasattr(target, 'after_update'):
                    target.after_update(mapper, connection)

            @event.listens_for(cls, 'before_delete')
            def before_delete(mapper, connection, target):
                if hasattr(target, 'before_delete'):
                    target.before_delete(mapper, connection)

            @event.listens_for(cls, 'after_delete')
            def after_delete(mapper, connection, target):
                if hasattr(target, 'after_delete'):
                    target.after_delete(mapper, connection)

    return BaseModel



from enum import Enum
from typing import List, Union, Any

from wedeliver_core_plus.helpers.service_config import ServiceConfig


DYNAMIC_LOOKUP_OPERATORS = ("=", "!=")
STATIC_LOOKUP_OPERATORS = (*DYNAMIC_LOOKUP_OPERATORS, "%LIKE%", "%LIKE", "LIKE%", ">=", "<=", ">", "<")


class Service:
    CAPTAIN = ServiceConfig('CAPTAIN_SERVICE')
    FINANCE = ServiceConfig('FINANCE_SERVICE')
    SDD = ServiceConfig('SDD_SERVICE')
    SUPPLIER = ServiceConfig('SUPPLIER_SERVICE')
    PN = ServiceConfig('PN_SERVICE')
    FINTECH = ServiceConfig('FINTECH_SERVICE')
    STC = ServiceConfig('STC_SERVICE')
    AUTH = ServiceConfig('AUTH_SERVICE')
    MAIL = ServiceConfig('MAIL_SERVICE')
    SMS = ServiceConfig('SMS_SERVICE')
    APILAYER = ServiceConfig('APILAYER_SERVICE')
    INVOICE = ServiceConfig('INVOICE_SERVICE')
    ADDRESS = ServiceConfig('ADDRESS_SERVICE')
    PUBLIC = ServiceConfig('PUBLIC_SERVICE')
    INTERNAL_NOTIFICATION = ServiceConfig('INTERNAL_NOTIFICATION_SERVICE')


class QueryTypes(Enum):
    SIMPLE_TABLE = 1
    FUNCTION = 2
    SEARCH = 3
    ADVANCED_TABLE = 4
    POPULATE_TABLE = 5


class LookUpType(Enum):
    DYNAMIC = "dynamic"
    STATIC = "static"


class FilterOperatorsType(Enum):
    AND = "AND"
    OR = "OR"


class InstallmentType(Enum):
    LEASE = 'Lease'
    PERSONAL_LOAN = 'Personal Loan'


def list_enum_values(enum_type):
    "return list of dict the enum values and names for the given enum type"
    return [dict(id=e.value, Value=e.name.capitalize()) for e in enum_type]


def list_enum_values_v2(enum_type):
    "return list of dict the enum values and names for the given enum type"
    return [dict(id=e.value, value=e.value) for e in enum_type]


def get_enum_value(enum_value, enum_type):
    "return  dict the enum values and names for the given enum type"
    for e in enum_type:
        if e.value == enum_value:
            return dict(id=e.value, Value=e.name.replace("_", " ").capitalize())


def format_values_dict(**values):
    "return  dict the enum values and names for the given enum type"
    return dict(**values)


class OrderByEnum(Enum):
    asc = "asc"
    desc = "desc"


def format_engine_size(engine_size_enum_type):
    "return the engine size in the format of 1.0 L"
    return [
        dict(id=e.value, Value=e.name[1:].replace("_", ".").capitalize() + " L")
        for e in engine_size_enum_type
    ]


def format_enum_with_dash(engine_size_enum_type):
    "return the enum containing dash to be without dash"
    return [
        dict(id=e.value, Value=e.name.replace("_", " ").capitalize())
        for e in engine_size_enum_type
    ]


class GetEumValue(Enum):
    @classmethod
    def get_value_of_key(cls, color_name):
        if color_name is None:
            return None

        member = cls.__members__.get(color_name.lower())
        if member is not None:
            return member.value  # Return the value of the member
        return None

    @classmethod
    def get_key_of_value(cls, value):
        if value is None:
            return None

        for key, val in cls.__members__.items():
            if val.value == value:
                return key.capitalize()  # Capitalize the first letter of the key

        return None


class VehicleColorEnum(GetEumValue):
    red = 1
    blue = 2
    black = 3
    silver = 4
    gray = 5
    white = 6
    maroon = 7
    purple = 9
    fuchsia = 10
    green = 11
    lime = 12
    olive = 13
    yellow = 14
    navy = 15
    teal = 17
    aqua = 18
    black_and_red = 19


class App(Enum):
    """Languages"""

    CUSTOMER_APP = "Customer App"
    WEB_APP = "Web App"
    UBER_LEAD_WEB_APP = "Uber Lead Web App"



from datetime import datetime, timedelta

import jwt
import pytest
import sqlalchemy

from app import app as flask_app
from app import db
import os
from wedeliver_core_plus import route_metadata, MockMicroFetcher, Role
from unittest.mock import patch
import re
from wedeliver_core_plus.helpers.enums import App


@pytest.fixture(scope='session')
def test_app():
    flask_app.config.update(
        {
            "TESTING": False,
            "DEBUG": False,
            "SQLALCHEMY_ECHO": False,
            "SQLALCHEMY_BINDS": None,
            "SQLALCHEMY_DATABASE_URI": "{engine}://{username}:{password}@{server}:{port}/{database}?charset=utf8".format(
                engine=os.environ.get("DATABASE_ENGINE", "mysql"),
                username=os.environ.get("DATABASE_USERNAME", "root"),
                password=os.environ.get("DATABASE_PASSWORD", "root"),
                server=os.environ.get("DATABASE_SERVER", "0.0.0.0"),
                database=os.environ.get("DATABASE_NAME", "test_db"),
                port=os.environ.get("DATABASE_PORT", 3306),
            ),
        }
    )
    with flask_app.app_context():
        # with flask_app.app_context(): ensures that all commands within this block are run in the Flask app context.
        db.create_all()
        # yield keyword returns a list of values but doesn't stop the execution of the function
        yield flask_app

        db.session.remove()
        # pytest.helpers.reset()
        db.drop_all()


@pytest.fixture()
def client(test_app):
    yield test_app.test_client()


# @pytest.fixture
# def db_session(test_app):
#     with test_app.app_context():
#         # Start a database transaction
#         connection = db.engine.connect()
#         transaction = connection.begin()
#         options = dict(bind=connection, binds={})
#         # Create a scoped session bound to the connection
#         session = db.create_scoped_session(options=options)
#
#         # Override the default db.session with our test session
#         prev_session = db.session  # Save the previous session
#         db.session = session
#
#         yield session
#
#         # Cleanup: Restore the original db.session
#         db.session = prev_session
#         session.remove()
#         transaction.rollback()
#         connection.close()


@pytest.fixture(scope='function')
def db_session(test_app):
    """
    Fixture to provide a database session with automatic transaction rollback.
    This will ensure all changes made during each test are rolled back,
    effectively deleting any rows inserted during the test.
    """
    with test_app.app_context():
        # Start a new transaction
        connection = db.engine.connect()
        transaction = connection.begin()

        # Bind a new session to the transaction connection
        options = {"bind": connection, "binds": {}}
        session = db.create_scoped_session(options=options)
        db.session = session

        yield session

        # Rollback the transaction to clean up inserted rows
        transaction.rollback()
        connection.close()

        # Remove session to clean up
        session.remove()


@pytest.fixture
def model_factory(db_session):
    """Generic factory fixture for creating models."""
    created_models = []

    def _create_model(create_func, **kwargs):
        model_instance = create_func(db_session, **kwargs)
        created_models.append(model_instance)
        return model_instance

    yield _create_model

    # Cleanup: Ensure models are deleted if not automatically cleaned up by rollback
    try:
        for model_instance in created_models:
            db_session.delete(model_instance)
        if created_models:
            db_session.commit()
    except Exception as e:
        pass
        # db_session.rollback()
        # pytest.fail(f"Unexpected error during cleanup: {e}")



# @pytest.fixture
# def model_factory(db_session):
#     def _create_model(create_func, **kwargs):
#         return create_func(db_session, **kwargs)
#     return _create_model


@pytest.fixture
def mock_micro_fetcher():
    # Define or import your MockMicroFetcher class or instance here
    # This should be the same as used in your other tests
    return MockMicroFetcher()


@pytest.fixture
def api_client(client, mock_micro_fetcher):
    class APIClient:
        def __init__(self, client):
            self.client = client
            self.default_headers = {}
            self.auth_token = None

        def set_headers(self, headers):
            self.default_headers.update(headers)

        def get(self, url, headers=None, **kwargs):
            return self._do_request(
                self.client.get,
                url,
                headers=headers,
                **kwargs
            )

        def post(self, url, json=None, data=None, headers=None, **kwargs):
            return self._do_request(
                self.client.post,
                url,
                json=json,
                data=data,
                headers=headers,
                **kwargs
            )

        def put(self, url, json=None, data=None, headers=None, **kwargs):
            return self._do_request(
                self.client.put,
                url,
                json=json,
                data=data,
                headers=headers,
                **kwargs
            )

        def delete(self, url, headers=None, **kwargs):
            return self._do_request(
                self.client.delete,
                url,
                headers=headers,
                **kwargs
            )

        def _merge_headers(self, headers):
            merged = self.default_headers.copy()
            if headers:
                merged.update(headers)
            return merged

        def _do_request(self, method, url, headers=None, **kwargs):
            merged_headers = self._merge_headers(headers)
            response = method(
                url,
                headers=merged_headers,
                **kwargs
            )
            return response

    return APIClient(client)


@pytest.fixture
def web_auth_token(test_app):
    def get_token(role=Role.SYSTEM_ADMIN, language='ar'):
        # Default token data for system admin
        token_data = dict(
            account_id=1,
            app=App.WEB_APP.value,
            email='administrator@wedeliverapp.com',
            role=role,
            is_logged=True,
            language=language,
            exp=int(datetime.timestamp(datetime.now() + timedelta(hours=12))),
            created_at=datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
        )

        # Encode the token
        token = jwt.encode(token_data, test_app.config.get("SECRET_KEY"), algorithm="HS256")

        plain_data = dict(
            email=token_data.get('email'),
            role=token_data.get("role"),
            language=language,
        )

        return dict(
            message="Logged successfully",
            data=plain_data,
            token=token.decode("utf-8") if isinstance(token, bytes) else token,
        )

    return get_token


@pytest.fixture
def web_supplier_token(test_app):
    def get_token(role=Role.CRM_USER, language='ar'):
        # Default token data for system supplier

        token_data = dict(
            account_id=1,
            app=App.WEB_APP.value,
            email='supplier@example.com',
            role=role,
            language=language,
            is_logged=True,
            exp=int(datetime.timestamp(datetime.now() + timedelta(hours=12))),
            created_at=datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            supplier_id=1
        )
        token = jwt.encode(token_data, test_app.config.get("SECRET_KEY"), algorithm="HS256")


        plain_data = dict(
            email=token_data.get("email"),
            role=token_data.get("role"),
            supplier="Test Supplier",
        )

        return dict(
            message="Logged successfully",
            data=plain_data,
            token=token.decode("utf-8") if isinstance(token, bytes) else token,
        )

    return get_token


@pytest.fixture
def app_auth_token(test_app):
    def get_token(role=Role.APP_USER, customer_id=1, account_id=2, device_id=1, language='ar', currency='SAR',
                  country_code='sa', nick_name='Test Nickname', image="",mobile= "966500000001",app_version="1.0.0",platform="android"):
        customer = dict(
            id=customer_id,
            nick_name=nick_name,
            image=image
        )
        token_data = dict(
            account_id=account_id,
            device_id=device_id,
            language=language,
            app=App.CUSTOMER_APP.value,
            role=role,
            customer=dict(**customer, full_name=customer.get("nick_name")),
            is_logged=True,
            exp=int(datetime.timestamp(datetime.now() + timedelta(days=365))),
            created_at=datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            mobile=mobile,
            app_version=app_version,
            platform=platform,
        )
        token = jwt.encode(token_data, test_app.config.get("SECRET_KEY"), algorithm="HS256")

        plain_data = dict(
            required_pin=False,
            language=token_data.get('language'),
            country_code=country_code,
            currency=currency,
            full_name=customer.get("nick_name"),
            customer_image=customer["image"],
            app_version_obj='0.0.0',
        )
        plain_data.update(**customer)

        # update_customer_last_login(account.id)
        return dict(
            message="Logged successfully",
            data=plain_data,
            token=token.decode("utf-8"),
        )

    return get_token


@pytest.fixture
def app_auth_pre_token(test_app):
    def get_token(role=Role.APP_USER, account_id=1):
        token_data = dict(
            account_id=account_id,
            app=App.CUSTOMER_APP.value,
            role=role,
            is_logged=False,
            exp=int(datetime.timestamp(datetime.now() + timedelta(days=365))),
            created_at=datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
        )
        token = jwt.encode(token_data, test_app.config.get("SECRET_KEY"), algorithm="HS256")

        db.session.commit()

        return dict(
            message="You have activated your account successfully",
            data=dict(required_pin=True),  # Require a (new PIN) or not
            token=token.decode("utf-8"),
        )

    return get_token


@pytest.fixture
def get_api_url():
    def _get_url(route_function_name):
        route_info = route_metadata.get(route_function_name)
        if not route_info:
            raise ValueError(f"No metadata found for route '{route_function_name}'")
        return route_info['url']

    return _get_url


@pytest.fixture
def auto_patch_microfetcher():
    patches = []

    def _patch(route_function_name, mock_micro_fetcher, manual=False):
        if manual:
            module_path = route_function_name
        else:
            route_info = route_metadata.get(route_function_name)
            if not route_info:
                raise ValueError(f"No metadata found for route '{route_function_name}'")

            if not route_info.get('is_testable'):
                raise ValueError(f"Route '{route_function_name}' is not testable")

            module_path = route_info.get('module_path')
            if not module_path:
                raise ValueError(f"No module path specified for route '{route_function_name}'")

        # Build the full module path to patch
        full_module_path = f"{module_path}.MicroFetcher"

        try:
            # Apply the patch
            patcher = patch(full_module_path, new=mock_micro_fetcher)
            patched = patcher.start()
            patches.append(patcher)
            return patched
        except (ModuleNotFoundError, AttributeError, ImportError) as e:
            # Module or MicroFetcher not found, skip patching
            # Optionally, log a warning message
            print(f"Warning: Could not patch MicroFetcher at '{full_module_path}'. Skipping patch.")
            print(f"Reason: {e}")
            return None

    yield _patch

    # Cleanup patches after tests
    for patcher in patches:
        patcher.stop()


@pytest.fixture
def get_route_metadata():
    return route_metadata


def substitute_url_placeholders(url, url_params):
    """
    Replaces placeholders in the URL with actual values from url_params.

    Placeholders are in the format: <converter:variable_name>
    Example: /path/<int:id> => /path/123

    Parameters:
    - url: The URL string with placeholders.
    - url_params: A dictionary of variable_name: value.

    Returns:
    - The URL with placeholders replaced by actual values.
    """

    def replace_match(match):
        # Extract variable name from the placeholder
        placeholder = match.group(0)  # e.g., '<int:withdraw_request_id>'
        # Split the placeholder to get the variable name
        parts = placeholder.strip('<>').split(':')
        if len(parts) == 2:
            variable_name = parts[1]
        else:
            variable_name = parts[0]
        # Get the actual value from url_params
        if variable_name in url_params:
            return str(url_params[variable_name])
        else:
            raise ValueError(f"URL parameter '{variable_name}' not provided in url_params")

    # Replace all placeholders in the URL
    return re.sub(r'<[^>]+>', replace_match, url)


@pytest.fixture
def api_request(api_client):
    def _api_request(route_function_name, url_params=None, method=None, **kwargs):
        """
        Makes an API request using the appropriate HTTP method based on route metadata.

        Parameters:
        - route_function_name: Name of the route function (string).
        - url_params: Dictionary of URL parameters to replace placeholders.
        - method: Optional HTTP method to use.
        - kwargs: Additional arguments for the API call, such as query_string, data, json, headers, etc.

        Returns:
        - response: The response object from the API call.
        """
        # Retrieve the route metadata
        route_info = route_metadata.get(route_function_name)
        if not route_info:
            raise ValueError(f"No metadata found for route '{route_function_name}'")

        # Get the API URL
        api_url = route_info.get('url')
        if not api_url:
            raise ValueError(f"No URL specified for route '{route_function_name}'")

        # Replace URL placeholders with actual values
        url_params = url_params or {}
        try:
            api_url = substitute_url_placeholders(api_url, url_params)
        except ValueError as e:
            raise ValueError(f"Error substituting URL placeholders: {e}")

        # Get the HTTP methods
        methods = route_info.get('methods', [])
        if not methods:
            raise ValueError(f"No HTTP methods specified for route '{route_function_name}'")

        # Choose the HTTP method to use
        if method:
            # Use the specified method
            http_method = method.lower()
            if http_method.upper() not in methods:
                raise ValueError(f"Specified method '{method.upper()}' not supported for route '{route_function_name}'")
        else:
            # Default to the first method in the list
            http_method = methods[0].lower()

        # Ensure the api_client has the method
        if not hasattr(api_client, http_method):
            raise AttributeError(f"api_client does not support method '{http_method}'")

        # Get the method from api_client
        api_client_method = getattr(api_client, http_method)

        # Make the API call
        response = api_client_method(api_url, **kwargs)

        return response

    return _api_request



from functools import wraps

from flask import request, jsonify, current_app

from wedeliver_core_plus.helpers.auth import (
    verify_user_token_v2,
    Auth,
    verify_user_token_v3,
    verify_user_token_v4,
    check_token_version,
    validate_device_fingerprint
)
from wedeliver_core_plus.helpers.exceptions import (
    AppValidationError,
    AppMissingAuthError, AppForbiddenError, AppDeprecatedApiError,
    AppExpiredTokenError, AppTokenRefreshRequired,
)
from flask_babel import _


def handle_auth(require_auth, append_auth_args=None, allowed_roles=None, pre_login=False,
                allowed_permissions=None, guards=[], deprecated=False):
    """
    A decorator for route functions that handles authentication with optional refresh token support.
    
    Args:
        require_auth: Whether authentication is required
        append_auth_args: List of arguments to extract from token and append to function kwargs
        allowed_roles: List of roles allowed to access this endpoint
        pre_login: Whether the endpoint is used in pre-login flow
        allowed_permissions: List of permissions allowed to access this endpoint
        guards: List of guard functions to check
        deprecated: Whether the endpoint is deprecated
    """

    def factory(func):
        @wraps(func)
        def inner_function(*args, **kws):
            if deprecated:
                raise AppDeprecatedApiError(_("This API is deprecated and no longer available"))

            if not require_auth:
                return func(*args, **kws)

            if "Authorization" not in request.headers:
                raise AppMissingAuthError(_("Missing authentication"))

            token = request.headers["Authorization"]

            if "country_code" not in request.headers and request.endpoint != "health_check":
                raise AppValidationError(_("Country Code is Required (c)"))

            # Get device fingerprint from headers if available
            device_fingerprint = request.headers.get("X-Device-Fingerprint")

            # Check token version and use appropriate verification method
            use_v4, version, token_payload = check_token_version(token,
                                                                 skip_expiration_check=True)

            if use_v4:
                if not device_fingerprint:
                    raise AppValidationError(_("Device fingerprint is required for token version 2 or higher"))

                # Validate device fingerprint matches the device_id in token
                if not validate_device_fingerprint(device_fingerprint, token_payload):
                    raise AppValidationError(_("Invalid device fingerprint or device ID mismatch"))

                user = verify_user_token_v4(token, device_fingerprint, skip_expiration_check=request.endpoint == "token_refresh_service")
            else:
                user = verify_user_token_v3(token=token)

            if not pre_login:
                if not user.get("is_logged"):
                    raise AppValidationError(_("Not Logged Token, please complete login process"))

            if allowed_roles:
                if user.get("role") not in allowed_roles:
                    raise AppValidationError(_("Not Allowed Role"))

            if guards:
                for guard in guards:
                    if not guard():
                        raise AppForbiddenError("Not Allowed Feature")

            if append_auth_args and isinstance(append_auth_args, list):
                for arg in append_auth_args:
                    if not kws.get('appended_kws'):
                        kws['appended_kws'] = dict()
                    if '.' in arg:
                        if 'as' in arg:
                            arg, as_arg = arg.split(' as ')
                        else:
                            as_arg = arg.replace('.', '_')

                        obj, key = arg.split('.')
                        value = user.get(obj, {}).get(key)
                        kws['appended_kws'][as_arg] = value
                    else:
                        value = user.get(arg)
                        kws['appended_kws'][arg] = value

            return func(*args, **kws)

        return inner_function

    return factory



from flask import request, g

from flask import session

from .authenticate import authenticate
from .micro_fetcher import MicroFetcher
from flask_babel import _

from wedeliver_core_plus.helpers.hmac_validator import verify_hmac_code
from wedeliver_core_plus.helpers.exceptions import AppExpiredTokenError, AppValidationError, AppTokenRefreshRequired
import jwt
import base64
import hmac
import hashlib
from wedeliver_core_plus import WedeliverCorePlus


class Auth:
    def __init__(self):
        pass

    @staticmethod
    def set_user(user):
        """
        Set user in flask global object and session
        """
        g.user = user  # store user in flask global object
        session["user"] = user

    @staticmethod
    def get_user():
        default_user_str = 'administrator@wedeliverapp.com'
        try:
            user = session.get("user", dict())
        except Exception:
            user = dict(user_id=default_user_str, email=default_user_str)

        return user

    @staticmethod
    def get_user_language():
        user = Auth.get_user()

        return user.get('language', 'en')

    @staticmethod
    def get_user_str():
        # app = WedeliverCorePlus.get_app()
        # with app.test_request_context():
        user = Auth.get_user()

        if user.get('email'):
            return user.get('email')
        else:
            return "Account-{}".format(
                user.get("account_id")
            )


def find_user_language(user=None):
    try:
        language = (
            request.headers["Accept-Language"].lower()
            if (
                    "Accept-Language" in request.headers
                    and request.headers["Accept-Language"] in ["en", "ar"]
            )
            else ((user.get("language") or 'ar') if user else 'ar')
        )
    except Exception:
        language = 'ar'

    return language


def validate_device_fingerprint(fingerprint, token_payload):
    """
    Validate if the given fingerprint matches the token's device_id.
    
    Args:
        fingerprint: The fingerprint to validate
        token_payload: The decoded token payload containing user and device information
        
    Returns:
        bool: True if fingerprint is valid and matches the token's device ID
    """
    if not fingerprint or not token_payload:
        return False
        
    try:
        # Check if it's a valid base64 encoded string
        decoded = base64.b64decode(fingerprint)
        # SHA256 HMAC should be 32 bytes
        if len(decoded) != 32:
            return False
            
        # Get account_id and device_id from token
        account_id = token_payload.get('account_id')
        device_id = token_payload.get('device_id')
        
        if not account_id or not device_id:
            return False
            
        # Verify the fingerprint against the expected one
        return verify_hmac_code(account_id, device_id, fingerprint)
        
    except Exception:
        return False


def check_token_version(token, skip_expiration_check=False):
    """
    Check token version and determine which verification method to use.
    
    Args:
        token: Auth token to check
        skip_expiration_check: If True, don't validate token expiration (default: False)
        
    Returns:
        tuple: (use_v4, version, token_payload) where use_v4 is a boolean indicating if v4 verification should be used,
               version is the token version (or None if not present), and token_payload is the decoded token data
    """
    try:
        # Assuming token is in format "Bearer <actual_token>"
        if token.startswith("Bearer "):
            token_value = token.split(" ")[1]
        else:
            token_value = token
            
        # Decode JWT token using the same method as in authenticate
        app = WedeliverCorePlus.get_app()
        verify_token = app.config.get("FLASK_ENV") == "production" or app.config.get("TESTING") == True
        
        # Decode without verification to read the payload
        # (we're just checking the version, not authenticating)
        options = {
            "verify_signature": verify_token
        }
        
        # Skip expiration check if requested
        if skip_expiration_check:
            options["verify_exp"] = False
            
        token_payload = jwt.decode(
            token_value,
            key=app.config.get("SECRET_KEY"),
            algorithms=["HS256"],
            options=options
        )
        
        # Check if version exists and is >= 1
        version = token_payload.get('token_version')
        if version is not None:
            return (int(version) >= 2, version, token_payload)
            
    except (jwt.DecodeError, jwt.ExpiredSignatureError, Exception) as e:
        # If any error occurs during parsing, default to v3
        print(f"Token parsing error: {str(e)}")
        pass
    
    return (False, None, None)


def verify_user_token_v2(token):
    results = MicroFetcher(
        "AUTH_SERVICE"
    ).from_function(
        "app.business_logic.auth.authenticate.authenticate"
    ).with_params(
        token=token
    ).fetch()

    results["data"].update(token=token)

    user = results["data"]
    user["language"] = find_user_language(user)

    Auth.set_user(user)

    return user


def verify_user_token_v3(token):
    results = authenticate(token)

    results["data"].update(token=token)
    user = results["data"]

    # get language form accept language in header of request
    user["language"] = find_user_language(user)

    Auth.set_user(user)

    return user


def verify_user_token_v4(token, device_fingerprint=None, skip_expiration_check=False):
    """
    Verify a user token with refresh token support.
    
    Args:
        token: The access token to verify
        device_fingerprint: Optional device fingerprint for verification
        skip_expiration_check: Optional skip expiration check

    Returns:
        dict: The decoded token data or new access token data
        
    Raises:
        AppExpiredTokenError: If the token is expired and no valid refresh token is available
        AppValidationError: If the token is invalid or refresh token verification fails
        AppTokenRefreshRequired: If the token needs to be refreshed and a new access token is available
    """
    try:
        # Try to verify the access token using auth service
        return MicroFetcher("AUTH_SERVICE").from_function(
            "app.business_logic.auth.mobile.token_operations.verify_access_token"
        ).with_params(
            token=token,
            skip_expiration_check=skip_expiration_check
        ).execute()

    except AppExpiredTokenError:
        # Access token has expired, check if refresh token is provided
        refresh_token = request.headers.get("X-Refresh-Token")
        
        if not refresh_token:
            # No refresh token, re-authentication is needed
            raise AppExpiredTokenError(_("Token expired, re-authentication required"))
        
        try:
            # Try to refresh the access token using auth service
            new_access_token = MicroFetcher("AUTH_SERVICE").from_function(
                "app.business_logic.auth.mobile.token_operations.refresh_access_token"
            ).with_params(
                refresh_token=refresh_token,
                device_fingerprint=device_fingerprint
            ).execute()
            
            # Raise specialized exception with the new access token
            raise AppTokenRefreshRequired(
                message=_("Token refreshed"),
                new_access_token=new_access_token
            )
            
        except (AppExpiredTokenError, AppValidationError):
            # Refresh token is also invalid or expired, re-authentication is needed
            raise AppExpiredTokenError(_("Session expired, re-authentication required"))



import json
from functools import wraps
from flask import Response


def handle_response(func):
    @wraps(func)
    def inner_function(*args, **kwargs):
        result = func(*args, **kwargs)
       
        if isinstance(result, Response):
            return result

        response = result
        code = 200
        use_default_response_message_key = True

        if isinstance(result, tuple):
            response = result[0]
            code = result[1]
            use_default_response_message_key = result[2]

        if code not in [200, 201, 204] and use_default_response_message_key:
            response = dict(message=response)

        return Response(
            json.dumps(response), content_type="application/json", status=code
        )

    return inner_function



class Topics:
    _instance = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super(Topics, cls).__new__(cls, *args, **kwargs)
            cls._instance._initialize()
        return cls._instance

    def _initialize(self):
        self.INTERNAL_NOTIFICATION_MESSAGE = 'internal_notification_message'
        self.SEND_MAIL = 'send_mail'
        self.SEND_SMS = 'send_sms'
        self.LOG_MODEL_CHANGES = 'log_model_changes'
        self.SEND_PUSH_NOTIFICATION = 'send_push_notification'
        self.THIRD_PARTY_ADD_BULK_TRANSACTIONS = 'third_party_add_bulk_transactions'


from wedeliver_core_plus.helpers.topics import Topics
from wedeliver_core_plus.helpers import kafka_producers
from wedeliver_core_plus import WedeliverCorePlus


def send_sms(message, mobile):
    if not isinstance(mobile, list):
        mobile = [mobile]

    app = WedeliverCorePlus.get_app()
    sms_sender_name = app.config.get("SMS_SENDER_NAME")

    kafka_producers.send_topic(topic=Topics().SEND_SMS, datajson=dict(
        function_name='app.business_logic.bulk_sms.send_v2.send_bulk_v2',
        function_params=dict(
            sms_message=message, mobile_numbers=mobile, issuer=sms_sender_name
        )
    ))



def get_obj_value(_obj, key):
    if hasattr(_obj, '__dict__'):  # object
        return _obj.__dict__.get(key)

    elif isinstance(_obj, dict):  # dict
        return _obj.get(key)

    elif hasattr(_obj, key):  # class
        return getattr(_obj, key)


def get_obj_path_value(obj, path, default=None):
    """Fetches value at the specified path, handling mixed object and dict structures."""
    keys = path.split(".") if isinstance(path, str) else path

    current = obj
    for key in keys:
        current = get_obj_value(current, key)
        if current is None:  # Stop if any key along the path doesn't exist
            return default
    return current


def set_obj_value(_obj, attr_name, attr_value, append_if_exists=False):
    if not append_if_exists:
        try:
            setattr(_obj, attr_name, attr_value)
        except AttributeError:
            _obj[attr_name] = attr_value
    else:
        try:
            old_value = getattr(_obj, attr_name) if hasattr(_obj, attr_name) else None
            if old_value:
                if isinstance(old_value, dict):
                    setattr(_obj, attr_name, [old_value])
                getattr(_obj, attr_name).append(attr_value)
            else:
                setattr(_obj, attr_name, attr_value)
        except AttributeError:
            old_value = _obj.get(attr_name)
            if old_value:
                if isinstance(old_value, dict):
                    _obj[attr_name] = [old_value]

                _obj[attr_name].append(attr_value)
            else:
                _obj[attr_name] = attr_value



def fix_number(number):
    try:
        return round(float(number), 2)
    except Exception:
        pass

    return 0


def format_number(number):
    try:
        # Convert to float and round to 2 decimal places
        fixed_number = round(float(number), 2)
    except (ValueError, TypeError):
        # Handle invalid inputs by setting fixed_number to 0.00
        fixed_number = 0.00

    # Convert -0.00 to 0.00
    if fixed_number == -0.0:
        fixed_number = 0.0

    # Return the fixed number formatted to 2 decimal places
    return "{:.2f}".format(fixed_number)



# -*- coding: utf-8 -*-

import os
import json

from dotenv import load_dotenv
from flask_sqlalchemy import SQLAlchemy
from sqlalchemy.orm import Session, sessionmaker
from wedeliver_core_plus import WedeliverCorePlus


class RoutingSession(Session):
    # Declare app and db as class attributes (they will be set only once)
    db = None
    app = None

    def get_bind(self, mapper=None, clause=None, **kwargs):
        if not self.app:
            self.app = WedeliverCorePlus.get_app()
            self.db = self.app.extensions['sqlalchemy'].db
        # If the clause is a SELECT, route to the read replica.
        if clause is not None and clause.__visit_name__ == 'select':
            self.app.logger.debug("Call read-database-replica")
            return self.db.get_engine(bind='read')
        # Otherwise, default to the write engine.
        self.app.logger.debug("Call write-database")
        return self.db.get_engine()


class CustomSQLAlchemy(SQLAlchemy):
    def create_session(self, options):
        # Remove any 'class_' in options to avoid duplicates
        options.pop('class_', None)
        # Use your custom session class instead of the default SignallingSession
        return sessionmaker(class_=RoutingSession, **options)


class Config(object):
    SQLALCHEMY_DATABASE_URI = None
    SQLALCHEMY_BINDS = None

    def init_database_config(self):
        def build_uri(server):
            return (
                f"{os.environ.get('DATABASE_ENGINE')}://"
                f"{os.environ.get('DATABASE_USERNAME')}:"
                f"{os.environ.get('DATABASE_PASSWORD')}@"
                f"{server}:"
                f"{os.environ.get('DATABASE_PORT', 3306)}/"
                f"{os.environ.get('DATABASE_NAME')}?charset=utf8"
            )

        primary_server = os.environ.get("DATABASE_SERVER")
        readonly_server = os.environ.get("DATABASE_READONLY")

        self.SQLALCHEMY_DATABASE_URI = build_uri(primary_server)
        self.SQLALCHEMY_BINDS = {
            "read": build_uri(readonly_server) if readonly_server else self.SQLALCHEMY_DATABASE_URI
        }

    def __init__(self):
        if not os.environ.get("DATABASE_ENGINE"):
            load_dotenv()

        self.init_database_config()

        for k, v in os.environ.items():

            if not k:
                continue

            env_val = os.environ.get(k).strip()
            # Convert Null and empty value to None
            env_val = None if env_val.upper() in ["NULL", ""] else env_val
            # Convert True and False value to Boolean
            env_val = (
                json.loads(env_val.lower())
                if env_val and env_val.upper() in ["TRUE", "FALSE"]
                else env_val
            )

            self.__dict__[k] = env_val



from wedeliver_core_plus import WedeliverCorePlus


class Transactions(object):
    savepoint = None
    db = None

    def __init__(self):
        app = WedeliverCorePlus.get_app()

        self.db = app.extensions['sqlalchemy'].db
        self.savepoint = self.db.session.begin_nested()

    @staticmethod
    def atomic():
        return Transactions()

    def __enter__(self):
        return self

    def __exit__(self, a, error, traceback):
        if error:
            if self.savepoint.is_active:
                self.savepoint.rollback()

            self.db.session.rollback()
            raise error

        self.db.session.commit()

    def commit(self, instance):
        self.db.session.add(instance)
        if not self.savepoint.is_active:
            self.savepoint = self.db.session.begin_nested()

        self.savepoint.commit()



from wedeliver_core_plus.helpers.sql import sql
import importlib
import os

from wedeliver_core_plus.helpers.enums import QueryTypes
from wedeliver_core_plus.helpers.search_function import search_function
from datetime import datetime,date

def fetch_relational_data(
        fields=None,
        table_name=None,
        column_name=None,
        compair_operator=None,
        column_values=None,
        functions=None,
        query_type=None,
        search_list=None,
        append_extra=None,
        use_country_code=None,
        filters_statements=None,
):
    # app = WedeliverCorePlus.get_app()
    # db = app.extensions['sqlalchemy'].db

    if query_type == QueryTypes.SEARCH.value:
        result, validation = search_function(table_name=table_name, search_list=search_list, append_extra=append_extra,
                                             use_country_code=use_country_code)
        return dict(
            result=result,
            validation=validation
        )
    else:
        relational_data_result = []

        if table_name:
            query = """
                SELECT {fields}
                FROM {table_name}
                WHERE {column_name} {compair_operator} {column_values}
                """.format(
                fields=', '.join(fields),
                table_name=table_name,
                column_name=column_name,
                compair_operator=compair_operator,
                column_values="('{}')".format(
                    "', '".join([str(val) for val in column_values])) if compair_operator == 'IN' else "'{}'".format(
                    column_values),
            )

            if filters_statements:
                filters_query = _formate_filters_query(filters_statements)
                query += f"AND {filters_query}"
            relational_data_result = sql(query)
            # check if the result contain datetime, or date object and convert it to string
            for row in relational_data_result:
                for key, value in row.items():
                    if isinstance(value, datetime):
                        row[key] = value.strftime('%Y-%m-%d %H:%M:%S')
                    elif isinstance(value, date):
                        row[key] = value.strftime('%Y-%m-%d')

        if functions and isinstance(functions, list):
            def _update_relational_data_result(function_result, fields):
                if relational_data_result:
                    for rd in relational_data_result:
                        for f in fields:
                            rd[f] = function_result.get(f) if isinstance(function_result,
                                                                         dict) else function_result.__dict__.get(f)
                else:
                    fields_dict = dict()
                    for f in fields:
                        fields_dict[f] = function_result.get(f) if isinstance(function_result,
                                                                              dict) else function_result.__dict__.get(f)
                    relational_data_result.append(fields_dict)

            for func in functions:
                function_file, function_call = os.path.splitext(func.get('name'))
                m = importlib.import_module(function_file)
                method = getattr(m, function_call[1:])

                function_result = method(**func.get('params'))

                fields = func.get('fields')
                if not fields:
                    relational_data_result = function_result
                elif function_result:
                    if isinstance(function_result, list):
                        for row in function_result:
                            _update_relational_data_result(row, fields)
                    else:
                        _update_relational_data_result(function_result, fields)

    return relational_data_result

def _formate_filters_query(filters_statements):
    filter_query = "("
    operator = filters_statements.get("operator")
    entries = filters_statements.get("entries", [])
    entry_index = 0
    for entry in entries:
        column_name = entry.get("column_name")
        compair_operator = entry.get("compair_operator")
        column_values = entry.get("column_values")
        if compair_operator in ('IN', 'NOT IN'):
            column_values = f"({', '.join([str(val) for val in column_values])})"
        else:
            column_values = f"'{column_values}'"
        if entry_index != 0:
            filter_query += f" {operator} "
        filter_query += f"{column_name} {compair_operator} {column_values}"
        entry_index += 1

    children = filters_statements.get("children", [])
    if children:
        for child in children:
            if entry_index != 0:
                filter_query += f" {operator} "
            filter_query += f"{_formate_filters_query(child)}"
            entry_index += 1
    filter_query += ")"
    return filter_query
